class NotificationEvent:
    """Notification Events."""

    FAILED = 'failed'
    ACCEPTED = 'accepted'
    VALIDATED = 'validated'
    AUTHORIZED = 'authorized'
    DISPATCHED = 'dispatched'
    RECEIVED = 'received'
    CONSUMED = 'consumed'



from typing import Any
from .envelope import Envelope
from .reason import Reason


class Command(Envelope):
    """Command representation."""

    def __init__(
        self,
        method: str,
        uri: str = None,
        type_n: str = None,
        resource: Any = None,
        status: str = None,
        reason: Reason = None,
        timeout: bool = None
    ) -> None:
        super().__init__()
        self.method = method
        self.uri = uri
        self.type_n = type_n
        self.resource = resource
        self.status = status
        self.reason = reason
        self.timeout = timeout



from ..constants import AuthenticationScheme
from .authentication import Authentication


class ExternalAuthentication(Authentication):
    """External authentication representation."""

    def __init__(self, token: str, issuer: str) -> None:
        super().__init__(AuthenticationScheme.EXTERNAL)
        self.token = token
        self.issuer = issuer



from .authentication_scheme import *
from .command_method import *
from .command_status import *
from .common_constants import *
from .content_types import *
from .notification_event import *
from .reason_code import *
from .session_compression import *
from .session_encryption import *
from .session_state import *
from .uri_templates import *



from asyncio import Future, get_running_loop

from ..command import Command
from ..constants import SessionState
from ..message import Message
from ..notification import Notification
from ..security import Authentication
from ..session import Session
from .channel import Channel


class ClientChannel(Channel):
    """Client channel representation."""

    async def establish_session_async(
        self,
        compression: str,
        encryption: str,
        identity: str,
        authentication: Authentication,
        instance: str
    ) -> Session:
        """Esablish a new session.

        Args:
            compression (str): compression type
            encryption (str): encryption type
            identity (str): identity str
            authentication (Authentication): Authentication specs
            instance (str): instance name

        Returns:
            Session: A established Session
        """
        self.ensure_state([SessionState.NEW], True)

        session: Session = await self.start_new_session_async()
        if session.encryption_options or session.compression_options:
            compression = compression if compression else session.compression_options[0]  # noqa: E501
            encryption = encryption if encryption else session.encryption_options[0]  # noqa: E501
            session: Session = await self.negotiate_session_async(compression, encryption)  # noqa: E501
        else:
            if session.compression != self.transport.compression:
                self.transport.set_compression(session.compression)
            if session.encryption != self.transport.encryption:
                self.transport.set_encryption(session.encryption)

        session: Session = await self.authenticate_session_async(identity, authentication, instance)  # noqa: E501
        self.__reset_session_listeners()
        return session

    def start_new_session_async(self) -> Future:
        """Start new session.

        Returns:
            Future: A new Session
        """
        self.ensure_state([SessionState.NEW], True)

        loop = get_running_loop()
        future = loop.create_future()

        self.on_session_negotiating = future.set_result
        self.on_session_authenticating = future.set_result
        self.__on_session_failed = future.set_exception

        session = Session(SessionState.NEW)
        self.send_session(session)

        return future

    def negotiate_session_async(
        self,
        session_compression: str,
        session_encryption: str
    ) -> Future:
        """Handle session in negotiating state.

        Args:
            session_compression (str): session compression type
            session_encryption (str): session encryption type

        Returns:
            Future: A negotiated Session
        """
        self.ensure_state([SessionState.NEGOTIATING], True)

        loop = get_running_loop()
        future = loop.create_future()

        self.on_session_authenticating = future.set_result
        self.__on_session_failed = future.set_exception

        session = Session(
            SessionState.NEGOTIATING,
            encryption=session_encryption,
            compression=session_compression
        )

        session.id = self.session_id
        self.send_session(session)

        return future

    def authenticate_session_async(
        self,
        identity: str,
        authentication: Authentication,
        instance: str
    ) -> Future:
        """Authenticate session.

        Args:
            identity (str): Identity to authenticate
            authentication (Authentication): Authentication object
            instance (str): Instance to authenticate

        Returns:
            Future: An authenticated Session
        """
        self.ensure_state([SessionState.AUTHENTICATING], True)

        loop = get_running_loop()
        future = loop.create_future()

        self.on_session_established = future.set_result
        self.__on_session_failed = future.set_exception

        session = Session(
            SessionState.AUTHENTICATING,
            scheme=authentication.scheme if authentication.scheme else 'unknown',  # noqa: E501
            authentication=authentication
        )
        session.from_n = f'{identity}/{instance}'
        session.id = self.session_id
        self.send_session(session)

        return future

    def send_finishing_session_async(self) -> Future:
        """Handle session in state finishing.

        Returns:
            Future: session future
        """
        self.ensure_state([SessionState.ESTABLISHED], True)

        loop = get_running_loop()
        future = loop.create_future()

        self.__on_session_finished = future.set_result
        self.__on_session_failed = future.set_exception

        session = Session(SessionState.FINISHING)
        session.id = self.session_id

        self.send_session(session)

        return future

    def on_session_finished(self, session: Session):
        """Handle callback on session finished.

        Args:
            session (Session): Received session
        """
        pass

    def on_session_failed(self, session: Session):
        """Handle callback on session failed.

        Args:
            session (Session): Received Session
        """
        pass

    def on_session(self, session: Session) -> None:  # noqa: WPS213
        """Handle session envelope received.

        Args:
            session (Session): Received Session
        """
        self.session_id = session.id
        self.state = session.state

        if session.state == SessionState.ESTABLISHED:
            self.local_node = str(session.to)
            self.remote_node = session.from_n

        # Switch case
        if session.state == SessionState.NEGOTIATING:
            self.on_session_negotiating(session)
            return

        if session.state == SessionState.AUTHENTICATING:
            self.on_session_authenticating(session)
            return

        if session.state == SessionState.ESTABLISHED:
            self.on_session_established(session)
            return

        if session.state == SessionState.FINISHED:
            self.transport.close()
            self.on_session_finished(session)
            self.__on_session_finished(session)
            return

        if session.state == SessionState.FAILED:
            self.transport.close()
            self.on_session_failed(session)
            self.__on_session_failed(session)
            return

    def on_message(self, message: Message) -> None:  # noqa: D102
        pass

    def on_command(self, command: Command) -> None:  # noqa: D102
        pass

    def on_notification(  # noqa: D102
        self,
        notification: Notification
    ) -> None:
        pass

    def on_session_authenticating(self, session: Session) -> None:
        """Handle session authenticating callback.

        Args:
            session (Session): received Session
        """
        pass

    def on_session_negotiating(self, session: Session) -> None:
        """Handle session negotiating callback.

        Args:
            session (Session): received Session
        """
        pass

    def on_session_established(self, session: Session) -> None:
        """Handle session established callback.

        Args:
            session (Session): received Session
        """
        pass

    def __reset_session_listeners(self) -> None:
        self.__on_session_finished = \
            self.on_session_negotiating = \
            self.on_session_established = \
            self.on_session_authenticating = \
            self.__on_session_failed = self.__empty_method  # noqa: WPS429

    def __on_session_failed(self, session: Session) -> None:
        pass

    def __on_session_finished(self, session: Session) -> None:
        pass

    def __empty_method(self) -> None:
        pass



from .command import *
from .constants import *
from .envelope import *
from .identity import *
from .listeners import *
from .network import *
from .node import *
from .notification import *
from .reason import *
from .security import *
from .serializable import *
from .client import *



from ..constants import AuthenticationScheme
from .authentication import Authentication


class GuestAuthentication(Authentication):
    """Guest authentication representation."""

    def __init__(self) -> None:
        super().__init__(AuthenticationScheme.GUEST)



from abc import abstractmethod

from ...listeners import MessageListener
from ...message import Message


class MessageChannel(MessageListener):
    """Message Channel."""

    @abstractmethod
    def send_message(self, message: Message) -> None:
        """Send a message.

        Args:
            message (Message): message to be sent
        """
        pass



class Reason:
    """A reason for events occurred during the client-server interactions."""

    def __init__(self, code: int, description: str) -> None:
        self.code = code
        self.description = description

    def __str__(self) -> str:
        """Override str representation.

        Returns:
            str: description with code
        """
        return f'{self.description} (Code {self.code.value})'



class CommandMethod:
    """Command methods."""

    GET = 'get'
    SET = 'set'
    DELETE = 'delete'
    MERGE = 'merge'
    OBSERVE = 'observe'
    SUBSCRIBE = 'subscribe'
    UNSUBSCRIBE = 'unsubscribe'



from asyncio import Future, get_running_loop, wait_for
from functools import partial
from typing import Callable, Dict, List

from ..command import Command
from ..constants import (CommandMethod, CommandStatus, ContentTypes,
                         NotificationEvent, SessionState, UriTemplates)
from ..envelope import Envelope
from ..message import Message
from ..network import Transport
from ..notification import Notification
from ..session import Session
from ..utilities import DictToClass
from .channels import (CommandChannel, MessageChannel, NotificationChannel,
                       SessionChannel)
from .command_processor import CommandProcessor


class Channel(
    MessageChannel,
    NotificationChannel,
    SessionChannel,
    CommandChannel,
    CommandProcessor
):
    """Channel to communicate with Envelopes."""

    def __init__(
        self,
        transport: Transport,
        auto_reply_pings: bool = True,
        auto_notify_receipt: bool = True
    ) -> None:
        self.transport = transport
        self.auto_reply_pings = auto_reply_pings
        self.auto_notify_receipt = auto_notify_receipt
        self.state = SessionState.NEW
        self.remote_node: str = None
        self.local_node: str = None
        self.session_id: str = None
        self.command_resolves: Dict[str, Callable] = {}
        self.transport.on_envelope = self.on_envelope

    def send_message(self, message: Message) -> None:  # noqa: D102
        self.__send_only_established(message)

    def send_command(self, command: Command) -> None:  # noqa: D102
        self.__send_only_established(command)

    def send_notification(self, notification: Notification) -> None:  # noqa: D102, E501
        self.__send_only_established(notification)

    def send_session(self, session: Session) -> None:  # noqa: D102
        self.ensure_state(
            [SessionState.FINISHED, SessionState.FAILED],
            False
        )
        self.__send(session)

    async def process_command_async(  # noqa: D102
        self,
        command: Command,
        timeout: float
    ) -> Command:
        loop = get_running_loop()
        future = loop.create_future()
        self.command_resolves[command.id] = future.set_result

        self.send_command(command)

        future.add_done_callback(
            partial(
                Channel.raise_command_timeout,
                command=command,
                command_resolves=self.command_resolves
            )
        )
        await wait_for(future, timeout)

        return future.result()

    @staticmethod
    def raise_command_timeout(
        fut: Future,
        command: Command,
        command_resolves: Dict[str, Callable]
    ) -> None:
        """Raise a command timeout.

        Args:
            fut (Future): the originator future
            command (Command): the timed out sent command
            command_resolves (Dict[str, Callable]): the dictionary of futures resolves

        Raises:
            TimeoutError: Command could not be processed in the given timeout
        """  # noqa: E501
        if fut.cancelled():
            command_resolve = command_resolves.get(command.id)
            if command_resolve is not None:
                del command_resolves[command.id]

            raise TimeoutError(
                f'The following command processing has timed out: {command}'
            )

    def on_envelope(self, envelope: dict) -> None:  # noqa: WPS231
        """Handle received envelope on transport.

        Args:
            envelope (dict): The received raw envelope
        """
        envelope: Envelope = DictToClass(envelope, Envelope)

        if Envelope.is_message(envelope):
            envelope.__class__ = Message
            self.__notify_message(envelope)
            self.on_message(envelope)

        elif Envelope.is_notification(envelope):
            envelope.__class__ = Notification
            self.on_notification(envelope)

        elif Envelope.is_session(envelope):
            envelope.__class__ = Session
            self.on_session(envelope)

        elif Envelope.is_command(envelope):
            envelope.__class__ = Command
            if hasattr(envelope, 'status') and envelope.status:
                set_result = self.command_resolves.get(envelope.id)

                if set_result:
                    set_result(envelope)
                    del self.command_resolves[envelope.id]
                    return
            if self.__should_reply_command(envelope):
                self.send_command(self.__create_ping_command_reply(envelope))
            self.on_command(envelope)

    def ensure_state(self, states: List[str], is_allowed: bool) -> None:
        """Check if class state is at specified states if it's allowed.

        Args:
            states (List[str]): List of states
            is_allowed (bool): if true check if state is at specified states\
            else if isn't

        Raises:
            ValueError: Raise if state doesn't match the is_allowed config
        """
        if (self.state in states) ^ is_allowed:
            raise ValueError(f'Cannot send in the {self.state} state')

    def __send_only_established(self, envelope: Envelope) -> None:
        self.ensure_state([SessionState.ESTABLISHED], True)
        self.__send(envelope)

    def __send(self, envelope: Envelope) -> None:
        self.transport.send(envelope.to_json())

    def __notify_message(self, message: Message) -> None:
        if self.__should_notify_message(message):
            notification = Notification(NotificationEvent.RECEIVED)
            notification.id = message.id
            notification.to = message.to
            self.send_notification(notification)

    def __should_notify_message(self, message: Message) -> bool:
        return self.auto_notify_receipt and\
            message.id is not None and\
            message.from_n is not None and\
            self.__is_for_me(message)

    def __is_for_me(self, envelope: Envelope) -> bool:
        return envelope.to is None or \
            envelope.to == self.local_node or \
            self.local_node.startswith(envelope.to)

    def __should_reply_command(self, command: Command) -> bool:
        return self.auto_reply_pings and \
            command.id and \
            command.uri == UriTemplates.PING and \
            command.method == CommandMethod.GET and \
            self.is_for_me(command)  # noqa: WPS222

    def __create_ping_command_reply(self, command: Command) -> Command:
        reply = Command(
            method=CommandMethod.GET,
            status=CommandStatus.SUCCESS,
            type_n=ContentTypes.PING,
            resource={},
        )
        reply.id = command.id
        reply.to = command.from_n
        return reply



import json
from typing import Any

from humps import camelize

PRIVATE_TOKEN = '_'  # noqa: S105
NODE_KEY_TOKEN = '_n'  # noqa: S105


class Serializable:
    """Serializable objects to json."""

    def __str__(self) -> str:
        """Override str representation.

        Returns:
            str: json representation as str
        """
        return json.dumps(self.to_json())

    def __eq__(self, obj: object) -> bool:
        """Override equal to compare serializated content.

        Args:
            obj (object): the object to compare

        Returns:
            bool
        """
        return isinstance(obj, Serializable) and \
            self.to_json() == obj.to_json()

    def __hash__(self) -> int:
        """Create a hash based on serializated content.

        Returns:
            int
        """
        return hash(str(self))

    def to_json(self) -> dict:
        """
        Transform class properties to json.

        Returns:
            dict: the class json representation without private properties
        """
        return {
            self.normalize_key(key): self.serialize_value(value)
            for key, value in self.__dict__.items()  # noqa: WPS110
            if self.__should_serialize_property(key, value)
        }

    def normalize_key(self, key: str) -> str:
        """
        Normalize a class property name.

        Args:
            key (str): property name

        Returns:
            str: the normalized key
        """
        return camelize(key.replace(NODE_KEY_TOKEN, str()))

    def serialize_value(self, value: Any) -> Any:
        """Serialize a value if it's Serializable.

        Args:
            value (Any): property value

        Returns:
            Any: the serialized property
        """
        return value.to_json() if isinstance(value, Serializable) else value

    def __should_serialize_property(self, key: str, value) -> bool:
        return not key.startswith(PRIVATE_TOKEN) and value is not None



from typing import List

from .envelope import Envelope
from .reason import Reason


class Session(Envelope):
    """Session representation."""

    def __init__(
        self,
        state: str,
        encryption_options: List[str] = None,
        encryption: str = None,
        compression_options: List[str] = None,
        compression: str = None,
        scheme: str = None,
        authentication=None,
        reason: Reason = None
    ) -> None:
        super().__init__()
        self.state = state
        self.encryption_options = encryption_options
        self.encryption = encryption
        self.compression_options = compression_options
        self.compression = compression
        self.scheme = scheme
        self.authentication = authentication
        self.reason = reason



from .constants import CommonConstants


class Node:
    """Node represents a identity with an instance."""

    def __init__(self, identity=None, instance: str = None) -> None:
        self.identity = identity
        self.instance = instance

    def to_identity(self):
        """Parse a identity into a Node.

        Returns:
            Identity: Return a identity object
        """
        return self.identity

    def __eq__(self, other) -> bool:
        """Override equality comparision.

        Args:
            other (any): other object

        Returns:
            bool: true if objects are equal.
        """
        return self.identity == other.identity \
            and self.instance == other.instance

    def __str__(self) -> str:
        """Represent Node as a string.

        Returns:
            str: Node as string
        """
        if self.instance is None:
            return str(self.identity)

        return f'{self.identity}/{self.instance}'

    @staticmethod
    def parse_str(possible_node: str):
        """Parse a string into a node.

        Args:
            possible_node (str): string node

        Returns:
            Node: Node object
        """
        from .identity import Identity  # noqa: WPS433

        props = possible_node.split(CommonConstants.DOMAIN_SEPARATOR)
        if len(props) < 2:
            return None
        name, instance_domain = props
        props = instance_domain.split(CommonConstants.INSTANCE_SEPARATOR)
        if len(props) == 2:
            domain, instance = props
            return Node(Identity(name, domain), instance)
        return Node(Identity(name, instance_domain))

    @staticmethod
    def parse(possible_node):
        """Parse a possible node into a Node.

        Args:
            possible_node (str | Identity| Node): possible Node object

        Returns:
            Node: Returns a node object
        """
        from .identity import Identity  # noqa: WPS433

        if isinstance(possible_node, Node):
            return possible_node
        elif isinstance(possible_node, Identity):
            return Node(possible_node)

        return Node.parse_str(possible_node)



class SessionState:
    """Session state."""

    NEW = 'new'
    NEGOTIATING = 'negotiating'
    AUTHENTICATING = 'authenticating'
    ESTABLISHED = 'established'
    FINISHING = 'finishing'
    FINISHED = 'finished'
    FAILED = 'failed'



from ..constants import AuthenticationScheme
from .authentication import Authentication


class TransportAuthentication(Authentication):
    """Transport authentication representation."""

    def __init__(self) -> None:
        super().__init__(AuthenticationScheme.TRANSPORT)



from abc import abstractmethod

from ...command import Command
from ...listeners import CommandListener


class CommandChannel(CommandListener):
    """Command Channel."""

    @abstractmethod
    def send_command(self, command: Command) -> None:
        """Send a Command.

        Args:
            command (Command): Command to be sent
        """
        pass



from ..constants import AuthenticationScheme
from .authentication import Authentication


class PlainAuthentication(Authentication):
    """Plain authentication representation."""

    def __init__(self, password: str) -> None:
        super().__init__(AuthenticationScheme.PLAIN)
        self.password = password



from abc import ABC, abstractmethod

from ..message import Message


class MessageListener(ABC):
    """Message listener callback."""

    @abstractmethod
    def on_message(self, message: Message) -> None:
        """Handle callback to handle a received Message.

        Args:
            message (Message): the received Message
        """
        pass



from abc import ABC, abstractmethod

from ..envelope import Envelope


class EnvelopeListener(ABC):
    """Envelope listener callback."""

    @abstractmethod
    def on_envelope(self, envelope: Envelope) -> None:
        """Handle callback to handle a received Envelope.

        Args:
            envelope (Envelope): the received Envelope
        """
        pass



from .authentication import *
from .external_authentication import *
from .guest_authentication import *
from .key_authentication import *
from .plain_authentication import *
from .transport_authentication import *



from .envelope import Envelope


class Message(Envelope):
    """Message representation."""

    def __init__(self, type_n: str, content) -> None:
        super().__init__()
        self.type_n = type_n
        self.content = content



from abc import ABC, abstractmethod

from ..session import Session


class SessionListener(ABC):
    """Session Listener callback."""

    @abstractmethod
    def on_session(self, session: Session) -> None:
        """Handle callback to handle a received session.

        Args:
            session (Session): the received Session
        """
        pass



from ..constants import AuthenticationScheme
from .authentication import Authentication


class KeyAuthentication(Authentication):
    """Key authentication representation."""

    def __init__(self, key: str) -> None:
        super().__init__(AuthenticationScheme.KEY)
        self.key = key



from typing import Dict

from .serializable import Serializable


class Envelope(Serializable):
    """Envelope representation."""

    def __init__(
        self,
        id: str = None,
        from_n: str = None,
        to: str = None,
        pp: str = None,
        metadata: Dict[str, str] = None
    ) -> None:
        self.id = id
        self.from_n = from_n
        self.to = to
        self.pp = pp
        self.metadata = metadata

    @staticmethod
    def is_message(envelope) -> bool:
        """Check if a envelope is a Message.

        Args:
            envelope (Envelope): the Envelope to be checked

        Returns:
            bool: True if the given Envelope is a Message
        """
        return hasattr(envelope, 'content')

    @staticmethod
    def is_notification(envelope) -> bool:
        """Check if a envelope is a Notification.

        Args:
            envelope (Envelope): the Envelope to be checked

        Returns:
            bool: True if the given Envelope is a Notification
        """
        return hasattr(envelope, 'event')

    @staticmethod
    def is_command(envelope) -> bool:
        """Check if a envelope is a Command.

        Args:
            envelope (Envelope): the Envelope to be checked

        Returns:
            bool: True if the given Envelope is a Command
        """
        return hasattr(envelope, 'method')

    @staticmethod
    def is_session(envelope) -> bool:
        """Check if a envelope is a Session.

        Args:
            envelope (Envelope): the Envelope to be checked

        Returns:
            bool: True if the given Envelope is a Session
        """
        return hasattr(envelope, 'state')



from abc import ABC, abstractmethod

from ..command import Command


class CommandListener(ABC):
    """Command listener callback."""

    @abstractmethod
    def on_command(self, command: Command) -> None:
        """Handle callback to handle a received command.

        Args:
            command (Command): the received Command
        """
        pass



from abc import abstractmethod

from ..command import Command
from ..listeners import CommandListener


class CommandProcessor(CommandListener):
    """Command Processor."""

    @abstractmethod
    async def process_command_async(
        self,
        command: Command,
        timeout: float
    ) -> Command:
        """Process a Command asynchronously and return the result.

        Args:
            command (Command): The Command to be processed
            timeout (float): Timeout to process the Command

        Returns:
            Command: The result Command
        """  # noqa: DAR202
        pass



from abc import abstractmethod
from typing import List

from ..envelope import Envelope
from ..listeners import EnvelopeListener


class Transport(EnvelopeListener):
    """Transport interface."""

    def __init__(self, compression: str, encryption: str) -> None:
        self.compression = compression
        self.encryption = encryption

    @abstractmethod
    def open(self, uri: str = None) -> None:
        """Open a new connection.

        Args:
            uri (str): the server uri
        """
        pass

    @abstractmethod
    def close(self) -> None:
        """Close a open connection."""
        pass

    @abstractmethod
    def send(self, envelope: Envelope) -> None:
        """Send a Envelope to the server.

        Args:
            envelope (Envelope): the Envelope to be sent
        """
        pass

    @abstractmethod
    def get_supported_compression(self) -> List[str]:
        """Get supported compressions by the server.

        Returns:
            list[SessionCompression]: a list of supported Compressions
        """  # noqa: DAR202
        pass

    @abstractmethod
    def set_compression(self, compression: str) -> None:
        """Set a compression to use with the server.

        Args:
            compression (SessionCompression): the compression to be used
        """  # noqa: DAR103
        pass

    @abstractmethod
    def get_supported_encryption(self) -> List[str]:
        """Get supported encryptions by the server.

        Returns:
            list[SessionEncryption]: a list of supported Encryption
        """  # noqa: DAR202
        pass

    @abstractmethod
    def set_encryption(self, encryption: str) -> None:
        """Set a encryption to use with the server.

        Args:
            encryption (SessionEncryption): the encryption to be used
        """  # noqa: DAR103
        pass



from os import path
from setuptools import find_packages, setup

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as file:
    long_description = file.read()

setup(
    name='lime_python',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    version='v2.0.0',
    license='apache-2.0',
    description='This module implements the LIME protocol',
    author='Gabriel Santos <gabrielr@take.net>, Luis Araújo <luisc@take.net>',
    author_email='gabrielr@take.net',
    url='https://github.com/chr0m1ng/lime-python',
    keywords=['lime', 'blip', 'chatbot'],
    install_requires=[
        'pyhumps'
    ],
    long_description=long_description,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9'
    ]
)



from abc import abstractmethod

from ...listeners import NotificationListener
from ...notification import Notification


class NotificationChannel(NotificationListener):
    """Notification Channel."""

    @abstractmethod
    def send_notification(self, notification: Notification) -> None:
        """Send a Notification.

        Args:
            notification (Notification): Notification to be sent
        """
        pass



class ReasonCode:
    """Default server reason codes."""

    # General error
    GENERAL_ERROR = 1
    # General session error
    SESSION_ERROR = 11
    # The session resource is already registered
    SESSION_REGISTRATION_ERROR = 12
    # An authentication error occurred
    SESSION_AUTHENTICATION_FAILED = 13
    # An error occurred while unregistering the session in the server
    SESSION_UNREGISTER_FAILED = 14
    # The required action is invalid for current session state
    SESSION_INVALID_ACTION_FOR_STATE = 15
    # The session negotiation has timed out
    SESSION_NEGOTIATION_TIMEOUT = 16
    # Invalid selected negotiation options
    SESSION_NEGOTIATION_INVALID_OPTIONS = 17
    # Invalid session mode requested
    SESSION_INVALID_SESSION_MODE_REQUESTED = 18
    # General validation error
    VALIDATION_ERROR = 21
    # The envelope document is null
    VALIDATION_EMPTY_DOCUMENT = 22
    # The envelope document MIME type is invalid
    VALIDATION_INVALID_RESOURCE = 23
    # The request status is invalid
    VALIDATION_INVALID_STATUS = 24
    # The request identity is invalid
    VALIDATION_INVALID_IDENTITY = 25
    # The envelope originator or destination is invalid
    VALIDATION_INVALID_RECIPIENTS = 26
    # The command method is invalid
    VALIDATION_INVALID_METHOD = 27
    # The command URI format is invalid
    VALIDATION_INVALID_URI = 27
    # General authorization error
    AUTHORIZATION_ERROR = 31
    # The sender is not authorized to send messages to the message destination
    AUTHORIZATION_UNAUTHORIZED_SENDER = 32
    # The destination doesn't have an active account
    AUTHORIZATION_DESTINATION_ACCOUNT_NOT_FOUND = 33
    # The envelope quota limit has been exceeded
    AUTHORIZATION_QUOTA_THRESHOLD_EXCEEDED = 34
    # General routing error
    ROUTING_ERROR = 41
    # The message destination was not found
    ROUTING_DESTINATION_NOT_FOUND = 42
    # The message destination gateway was not found
    ROUTING_GATEWAY_NOT_FOUND = 43
    # The message destination was not found
    ROUTING_ROUTE_NOT_FOUND = 44
    # General dispatching error
    DISPATCH_ERROR = 51
    # General command processing error
    COMMAND_PROCESSING_ERROR = 61
    # There's no command processor available for process the request
    COMMAND_RESOURCE_NOT_SUPPORTED = 62
    # The command method is not supported
    COMMAND_METHOD_NOT_SUPPORTED = 63
    # The command method has an invalid argument value
    COMMAND_INVALID_ARGUMENT = 64
    # The requested command is not valid for current session mode
    COMMAND_INVALID_SESSION_MODE = 65
    # The command method was not allowed
    COMMAND_NOT_ALLOWED = 66
    # The command resource was not found
    COMMAND_RESOURCE_NOT_FOUND = 67
    # General message processing error
    MESSAGE_PROCESSING_ERROR = 61
    # The message content type is not supported
    MESSAGE_UNSUPPORTED_CONTENT_TYPE = 71
    # General gateway processing error
    GATEWAY_ERROR = 81
    # The content type is not supported by the gateway
    GATEWAY_CONTENT_TYPE_NOT_SUPPORTED = 82
    # The message destination was not found on gateway
    GATEWAY_DESTINATION_NOT_FOUND = 83
    # The functionality is not supported by the gateway
    GATEWAY_NOT_SUPPORTED = 84
    # General application processing error
    APPLICATION_ERROR = 101



from abc import ABC, abstractmethod

from ..notification import Notification


class NotificationListener(ABC):
    """Notification listener callback."""

    @abstractmethod
    def on_notification(self, notification: Notification) -> None:
        """Handle callback to handle a received notification.

        Args:
            notification (Notification): the received Notification
        """
        pass



from .constants import CommonConstants


class Identity:
    """Represents an identity in a domain."""

    def __init__(self, name: str = None, domain: str = None) -> None:
        self.name = name
        self.domain = domain

    def __str__(self) -> str:
        """Override to str to reprensent a Identity.

        Returns:
            str: the str representation
        """
        return f'{self.name}@{self.domain}'

    def __eq__(self, other) -> bool:
        """Override equality comparision.

        Args:
            other (any): other object

        Returns:
            bool: true if objects are equal.
        """
        return self.name == other.name and self.domain == other.domain

    @staticmethod
    def parse_str(possible_identity: str):  # noqa: F821
        """Parse a str into a Identity.

        Args:
            possible_identity (str): the str to be parsed

        Returns:
            Identity: the created Identity
        """
        props = possible_identity.split(CommonConstants.DOMAIN_SEPARATOR)
        if len(props) < 2:
            return None
        name, domain = props
        return Identity(
            name,
            domain.split(CommonConstants.INSTANCE_SEPARATOR)[0]
        )

    @staticmethod
    def parse(possible_identity: str):
        """Parse a str | Identity | Node to a Identity.

        Args:
            possible_identity (str | Identity | Node): the value to be parsed

        Returns:
            Identity: the created Identity
        """  # noqa: DAR103
        from .node import Node  # noqa: WPS433

        if (isinstance(possible_identity, Identity)):
            return possible_identity
        elif (isinstance(possible_identity, Node)):
            return possible_identity.identity

        return Identity.parse_str(possible_identity)



from abc import abstractmethod

from ...listeners import SessionListener
from ...session import Session


class SessionChannel(SessionListener):
    """Session Channel."""

    @abstractmethod
    def send_session(self, session: Session) -> None:
        """Send a Session.

        Args:
            session (Session): Session to be sent
        """
        pass



from humps import decamelize

from ..serializable import Serializable

NORMALIZED_KEYS = frozenset(('from', 'type'))


class DictToClass(Serializable):
    """Turns a dictionary into a class."""

    def __init__(self, dictionary: dict, class_type: type) -> None:
        for key, value in dictionary.items():
            setattr(self, self.__normalize_key(key), value)
        self.__class__ = class_type

    def __normalize_key(self, key) -> str:
        key = decamelize(key)
        if key in NORMALIZED_KEYS:
            key = f'{key}_n'
        return key



from .envelope import Envelope
from .reason import Reason


class Notification(Envelope):
    """Notification representation."""

    def __init__(self, event: str, reason: Reason = None) -> None:
        super().__init__()
        self.event = event
        self.reason = reason
